import datetime
import json
import logging
import xml.etree.ElementTree as ET
from typing import Optional

from prpc_python import RpcApp

from cin_validator import cin_validator
from cin_validator.rules.ruleset_utils import get_year_ruleset

logger = logging.getLogger(__name__)
handler = logging.FileHandler(
    datetime.datetime.now().strftime("lac validator --%d-%m-%Y %H.%M.%S.log")
)

f_format = logging.Formatter("%(asctime)s - %(levelname)s - % (message)s")
handler.setFormatter(f_format)
logger.addHandler(handler)

app = RpcApp("validate_cin")


@app.call
def get_rules(collection_year: str) -> str:
    """
    :param str collection_year: validation year e.g "2023" for 2022/2023 validation rules.
    :return rules_df: available rule codes and definitions according to chosen ruleset.
    """
    ruleset_registry = get_year_ruleset(collection_year)

    rules = []
    for _, rule in ruleset_registry.items():
        rules.append(
            {
                "code": str(rule.code),
                "description": str(rule.code) + " - " + str(rule.message),
            }
        )

    return json.dumps(rules)


@app.call
def generate_tables(cin_data: dict) -> dict[str, dict]:
    """
    :param cin_data: files uploaded by user mapped to the field where files were uploaded.
    :return cin_data_tables:  a dictionary of dataframes that has been converted to json.
    """
    # Only a single XML file representing the current year is accepted as an input by the tool.
    cin_data_file = cin_data["This year"][0]
    filetext = cin_data_file.read().decode("utf-8")
    root = ET.fromstring(filetext)

    data_files = cin_validator.convert_data(root)

    # make data json-serialisable
    cin_data_tables = {
        table_name: table_df.to_json(orient="records")
        for table_name, table_df in data_files.items()
    }

    return cin_data_tables


@app.call
def cin_validate(
    cin_data: dict,
    file_metadata: dict,
    selected_rules: Optional[list[str]] = None,
):
    """
    :param cin_data: eys are table names and values are CIN csv files.
    :param file_metadata: contains collection year and local authority as strings.
    :param selected_rules: array of rules the user has chosen. consists of rule codes as strings.

    :return issue_report: issue locations in the data.
    :return rule_defs: codes and descriptions of the rules that triggers issues in the data.
    """
    cin_data_file = cin_data["This year"][0]
    filetext = cin_data_file.read().decode("utf-8")
    root = ET.fromstring(filetext)

    # fulltree = ET.parse("fake_data\\fake_CIN_data.xml")
    # root = fulltree.getroot()

    raw_data = cin_validator.convert_data(root)

    # Send string-format data to the frontend.
    cin_data_tables = {
        table_name: table_df.to_json(orient="records")
        for table_name, table_df in raw_data.items()
    }

    # Convert date columns to datetime format to enable comparison in rules.
    data_files = cin_validator.process_data(raw_data)
    # get rules to run based on specified year.
    ruleset_registry = get_year_ruleset(file_metadata["collectionYear"])

    # run validation
    validator = cin_validator.CinValidator(data_files, ruleset_registry, selected_rules)

    # make return data json-serialisable

    # what the frontend will display
    issue_report = validator.full_issue_df.to_json(orient="records")
    multichild_issues = validator.multichild_issues.to_json(orient="records")

    # what the user will download
    user_report = validator.user_report.to_json(orient="records")

    validation_results = {
        "issue_locations": [issue_report],
        "multichild_issues": [multichild_issues],
        "data_tables": [cin_data_tables],
        "user_report": [user_report],
    }
    return validation_results



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

Disabilities = CINTable.Disabilities
Disability = Disabilities.Disability


# define characteristics of rule
@rule_definition(
    code="8790",
    module=CINTable.Disabilities,
    message="Disability information includes both None and other values",
    affected_fields=[Disability],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Disabilities]
    """
    If a <Disability> (N00099) value of NONE is present, then there should be 
    no other <Disability> (N00099) values present for the same child
    """

    df.reset_index(inplace=True)
    df_orig = df.copy()
    # Form df containing only the NONE disability values
    df = df[df["Disability"].str.upper() == "NONE"]

    # Join back to the original dataframe via LAchildID
    df = df.merge(df_orig, how="left", on="LAchildID", suffixes=["", "_orig"])

    # Any disability values that aren't NONE in the merged table are now an error.
    df = df[df["Disability_orig"].str.upper() != "NONE"]

    # Return original index for the error rows
    failing_indices = df.set_index("index_orig").index

    rule_context.push_issue(table=Disabilities, field=Disability, row=failing_indices)


def test_validate():
    # 0      #1      #2      #3     #4      #5      #6      #7
    ids = ["1", "1", "2", "3", "3", "4", "4", "5"]
    dis_is = ["NONE", "AIND", "NONE", pd.NA, "MOTH", "NONE", "AAAA", "AA"]

    fake_dataframe = pd.DataFrame({"LAchildID": ids, "Disability": dis_is})

    result = run_rule(validate, {Disabilities: fake_dataframe})

    issues = list(result.issues)

    assert len(issues) == 2

    assert issues == [
        IssueLocator(CINTable.Disabilities, Disability, 1),
        IssueLocator(CINTable.Disabilities, Disability, 6),
    ]

    assert result.definition.code == "8790"
    assert (
        result.definition.message
        == "Disability information includes both None and other values"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate
Year = Header.Year


@rule_definition(
    code="100",
    module=CINTable.Header,
    message="Reference Date is incorrect",
    affected_fields=[ReferenceDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Header]

    # <ReferenceDate> (N00603) must be present and must equal 2022-03-31

    # ReferenceDate is expected to be the 31st of March in the collection year
    df["expected_date"] = "31/03/" + df[Year]
    df["expected_date"] = pd.to_datetime(
        df["expected_date"], format="%d/%m/%Y", errors="coerce"
    )

    # Checks that the reference date is present
    is_present = df[ReferenceDate].isna()
    # Checks the error date is equal to 31/03/[collection_year].
    error_date = df[ReferenceDate] != df["expected_date"]
    failing_indices = df[is_present | error_date].index

    rule_context.push_issue(table=Header, field=ReferenceDate, row=failing_indices)


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    # Sidenote: a typical Header table will only have one row.
    header = pd.DataFrame(
        [
            {Year: "2022", ReferenceDate: "31/03/2022"},
            {Year: "2022", ReferenceDate: pd.NA},
            {Year: "2022", ReferenceDate: pd.NA},
            {Year: "2022", ReferenceDate: "30/11/2021"},
        ]
    )

    header[ReferenceDate] = pd.to_datetime(
        header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(validate, {Header: header})

    issues = list(result.issues)
    # Intended fail points in data.
    assert len(issues) == 3
    # Intended failures of test data by index.
    assert issues == [
        IssueLocator(CINTable.Header, ReferenceDate, 1),
        IssueLocator(CINTable.Header, ReferenceDate, 2),
        IssueLocator(CINTable.Header, ReferenceDate, 3),
    ]

    # Checks rule code and message are correct.
    assert result.definition.code == "100"
    assert result.definition.message == "Reference Date is incorrect"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
LAchildID = ChildIdentifiers.LAchildID


@rule_definition(
    code="1580",
    module=CINTable.ChildIdentifiers,
    message="LA Child ID must not contain any non-alphanumeric characters",
    affected_fields=[LAchildID],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers].copy()
    non_alphanumeric = df[df["LAchildID"].str.isalnum() == False]

    failing_indices = non_alphanumeric.index

    rule_context.push_issue(
        table=ChildIdentifiers, field=LAchildID, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    # Sidenote: a typical Header table will only have one row.
    sample_ci = pd.DataFrame(
        {
            "LAchildID": [
                "XXXXXXXXXXXXXXXXXXXX",
                "XXXXXXXXXXXXXXXXXXX",
                "XXXXXXXXXXXXXXXXXXXX;",  # non-alphanumeric characters, fail
            ]
        }
    )

    result = run_rule(validate, {ChildIdentifiers: sample_ci})

    issues = list(result.issues)
    # Intended fail points in data.
    assert len(issues) == 1
    # Intended failures of test data by index.
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, LAchildID, 2),
    ]

    # Checks rule code and message are correct.
    assert result.definition.code == "1580"
    assert (
        result.definition.message
        == "LA Child ID must not contain any non-alphanumeric characters"
    )



"""
Rule number: 8530Q
Module: Child Identifiers
Rule details: If present <ExpectedPersonBirthDate> (N00098) should be between [<ReferenceDate> (N00603) minus 30 days] and [<ReferenceDate> (N00603) plus 9 months]
Rule message: Please check: Expected Date of Birth is outside the expected range for this census (March to December of the Census Year end)
"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
ExpectedPersonBirthDate = ChildIdentifiers.ExpectedPersonBirthDate
LAchildID = ChildIdentifiers.LAchildID

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="8530Q",
    module=CINTable.ChildIdentifiers,
    rule_type=RuleType.QUERY,
    message="Please check and either amend data or provide a reason: Expected Date of Birth is outside the expected range for this census (March to December of the Census Year end)",
    affected_fields=[ExpectedPersonBirthDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    df.index.name = "ROW_ID"

    df_ref = data_container[Header]
    ref_date = df_ref[ReferenceDate].iloc[0]

    #  <ExpectedPersonBirthDate> (N00098) should be between [<ReferenceDate> (N00603) minus 30 days] and [<ReferenceDate> (N00603) plus 9 months]

    # Filter to only those with expected birthdate
    df = df[~df[ExpectedPersonBirthDate].isna()]

    # Find the reference date - 30
    earliest_date = ref_date - pd.DateOffset(days=30)
    # Find reference date + 9 months
    latest_date = ref_date + pd.DateOffset(months=9)

    condition1 = df[ExpectedPersonBirthDate] >= latest_date
    condition2 = df[ExpectedPersonBirthDate] <= earliest_date

    df_issues = df[condition1 | condition2].reset_index()

    link_id = tuple(zip(df_issues[LAchildID], df_issues[ExpectedPersonBirthDate]))
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildIdentifiers,
        columns=[ExpectedPersonBirthDate],
        row_df=df_issues,
    )


def test_validate():
    sample_ChildIdentifiers = pd.DataFrame(
        [
            {
                "LAchildID": "ID1",
                "ExpectedPersonBirthDate": pd.NA,
                # Pass, no birth date
            },
            {
                "LAchildID": "ID2",
                "ExpectedPersonBirthDate": "30/03/2023",
                # Pass, birth date within range
            },
            {
                "LAchildID": "ID3",
                "ExpectedPersonBirthDate": "15/10/2021",
                # Fail, start date is before ref date - 45wd
            },
            {
                "LAchildID": "ID4",
                "ExpectedPersonBirthDate": "15/03/2024",
                # Fail, later than 9 months after ref date
            },
        ]
    )

    sample_ChildIdentifiers[ExpectedPersonBirthDate] = pd.to_datetime(
        sample_ChildIdentifiers[ExpectedPersonBirthDate],
        format="%d/%m/%Y",
        errors="coerce",
    )

    sample_header = pd.DataFrame([{"ReferenceDate": "31/03/2023"}])

    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_ChildIdentifiers,
            Header: sample_header,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 1
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns
    assert issue_columns == [ExpectedPersonBirthDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "ID3",
                    pd.to_datetime("15/10/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "ID4",
                    pd.to_datetime("15/03/2024", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
        ]
    )

    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8530Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: Expected Date of Birth is outside the expected range for this census (March to December of the Census Year end)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Section47 = CINTable.Section47
DateOfInitialCPC = Section47.DateOfInitialCPC
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8715",
    module=CINTable.Section47,
    message="Date of Initial Child Protection Conference must fall within the census year",
    affected_fields=[DateOfInitialCPC, ReferenceDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Section47]

    # ReferenceDate exists in the header table so we get header table too.
    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]
    # the make_census_period function generates the start and end date so that you don't have to do it each time.
    collection_start, reference_date = make_census_period(ref_date_series)

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # If present <DateOfInitialCPC> (N00110) must be on or between [Start_Of_Census_Year] and <ReferenceDate> (N00603)
    # A value is out of range if it is before the start or after the end.
    failing_indices = df[
        (df[DateOfInitialCPC] < collection_start)
        | (df[DateOfInitialCPC] > reference_date)
    ].index

    # Replace Section47 and DateOfInitialCPC with the table and column name concerned in your rule, respectively.
    rule_context.push_issue(
        table=Section47, field=DateOfInitialCPC, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]  # the census start date here will be 01/04/2021
    )
    fake_section47 = pd.DataFrame(
        [
            {
                DateOfInitialCPC: "01/03/2019"
            },  # 0 fail: March 1st is before April 1st, 2021. It is out of range
            {
                DateOfInitialCPC: "01/04/2021"
            },  # 1 pass: April 1st is within April 1st, 2021 to March 31st, 2022.
            {
                DateOfInitialCPC: "01/10/2022"
            },  # 2 fail: October 1st is after March 31st, 2022. It is out of range
        ]
    )

    # if date columns are involved, the validate function will be expecting them as dates so convert before passing them in.
    fake_section47[DateOfInitialCPC] = pd.to_datetime(
        fake_section47[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    # Since the ReferenceDate comes from the Header column, we provide that also.
    result = run_rule(validate, {Section47: fake_section47, Header: fake_header})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        # from above, index positions 0 and 2 fail.
        IssueLocator(CINTable.Section47, DateOfInitialCPC, 0),
        IssueLocator(CINTable.Section47, DateOfInitialCPC, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace '8715' with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8715"
    assert (
        result.definition.message
        == "Date of Initial Child Protection Conference must fall within the census year"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildProtectionPlans = CINTable.ChildProtectionPlans
LAchildID = ChildProtectionPlans.LAchildID
CPPID = ChildProtectionPlans.CPPID
CPPstartDate = ChildProtectionPlans.CPPstartDate
CPPendDate = ChildProtectionPlans.CPPendDate

CINplanDates = CINTable.CINplanDates
CINPlanStartDate = CINplanDates.CINPlanStartDate
CINPlanEndDate = CINplanDates.CINPlanEndDate

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="4017",
    module=CINTable.CINplanDates,
    message="A CIN Plan has been reported as open at the same time as a Child Protection Plan.",
    affected_fields=[
        CINPlanStartDate,
        CINPlanEndDate,
        CPPstartDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cpp = data_container[ChildProtectionPlans].copy()
    df_cin = data_container[CINplanDates].copy()

    df_cpp.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_cpp.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]

    collection_start, reference_date = make_census_period(ref_date_series)

    # The <CPPstartDate> (N00105) for any CPP group cannot fall within either:
    # <CINPlanStartDate> (N00689) or <CINPlanEndDate> (N00690);
    # or <CINPlanStartDate> and <ReferenceDate> (N00603) if <CINPlanEndDate> is not present - for any CIN Plan Group;
    # unless <CCPstartDate> is equal to <CINPlanEndDate> for this group.

    #  Merge tables
    df_merged = df_cpp.merge(
        df_cin,
        on=["LAchildID"],
        how="left",
        suffixes=("_cpp", "_cin"),
    )

    # Get rows where CPPstartDate is after CINPlanStartDate
    # and CPPstartDate before CINPlanEndDate (or if null, before/on ReferenceDate)
    cpp_start_after_cin_start = df_merged[CPPstartDate] >= df_merged[CINPlanStartDate]
    cpp_start_before_cin_end = (
        df_merged[CPPstartDate] < df_merged[CINPlanEndDate]
    ) & df_merged[CINPlanEndDate].notna()
    cpp_start_before_reference_date = (
        df_merged[CPPstartDate] <= reference_date
    ) & df_merged[CINPlanEndDate].isna()

    df_merged = df_merged[
        cpp_start_after_cin_start
        & (cpp_start_before_cin_end | cpp_start_before_reference_date)
    ].reset_index()

    df_merged["ERROR_ID"] = tuple(zip(df_merged[LAchildID], df_merged[CPPstartDate]))

    df_cpp_issues = (
        df_cpp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildProtectionPlans,
        columns=[CPPstartDate],
        row_df=df_cpp_issues,
    )
    rule_context.push_type_2(
        table=CINplanDates,
        columns=[CINPlanStartDate, CINPlanEndDate],
        row_df=df_cin_issues,
    )


def test_validate():
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2001"}]  # the census start date here will be 01/04/2000
    )

    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINPlanStartDate": "26/05/2000",
                "CINPlanEndDate": "30/05/2000",
            },
            {
                "LAchildID": "child1",
                "CINPlanStartDate": "26/06/2000",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "26/10/2000",
                "CINPlanEndDate": "10/12/2000",
            },
            {
                "LAchildID": "child2",
                "CINPlanStartDate": "26/02/2001",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child3",
                "CINPlanStartDate": "26/05/2000",
                "CINPlanEndDate": "30/10/2001",
            },
            {
                "LAchildID": "child5",
                "CINPlanStartDate": "26/05/2000",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child8",
                "CINPlanStartDate": "15/09/2000",
                "CINPlanEndDate": pd.NA,
            },
            {
                "LAchildID": "child8",
                "CINPlanStartDate": "09/06/2000",
                "CINPlanEndDate": "20/07/2000",
            },
        ]
    )
    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Pass - Before CIN
                "CPPstartDate": "04/04/2000",
            },
            {
                "LAchildID": "child1",  # 1 Fail - During CIN
                "CPPstartDate": "28/05/2000",
            },
            {
                "LAchildID": "child1",  # 2 Pass - Same as CIN End
                "CPPstartDate": "30/05/2000",
            },
            {
                "LAchildID": "child1",  # 3 Pass - Between CIN
                "CPPstartDate": "04/06/2000",
            },
            {
                "LAchildID": "child1",  # 4 Fail - During CIN (via reference_date)
                "CPPstartDate": "30/06/2000",
            },
            {
                "LAchildID": "child2",  # 5 Fail - Same as CIN Start
                "CPPstartDate": "26/10/2000",
            },
            {
                "LAchildID": "child2",  # 6 Fail - Same as CIN Start
                "CPPstartDate": "26/02/2001",
            },
            {
                "LAchildID": "child2",  # 7 Fail - During CIN (via reference_date)
                "CPPstartDate": "26/03/2001",
            },
            {
                "LAchildID": "child3",  # 8 Pass - Same as CIN End (future return year handled by different rule!)
                "CPPstartDate": "30/10/2001",
            },
            {
                "LAchildID": "child4",  # 9 Pass - No CIN
                "CPPstartDate": "04/06/2000",
            },
            {
                "LAchildID": "child5",  # 10 Fail - Start on ReferenceDate
                "CPPstartDate": "31/03/2001",
            },
            {
                "LAchildID": "child8",  # 11 Fail - between CinStartDate and CinEndDate
                "CPPstartDate": "22/06/2000",
            },
        ]
    )

    sample_cpp[CPPstartDate] = pd.to_datetime(
        sample_cpp[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINPlanStartDate] = pd.to_datetime(
        sample_cin[CINPlanStartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin[CINPlanEndDate] = pd.to_datetime(
        sample_cin[CINPlanEndDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            CINplanDates: sample_cin,
            Header: sample_header,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == ChildProtectionPlans

    issue_columns = issues.columns
    assert issue_columns == [CPPstartDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 7
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("28/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child1",
                    pd.to_datetime("30/06/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": (
                    "child2",
                    pd.to_datetime("26/10/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child2",
                    pd.to_datetime("26/02/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [6],
            },
            {
                "ERROR_ID": (
                    "child2",
                    pd.to_datetime("26/03/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [7],
            },
            {
                "ERROR_ID": (
                    "child5",
                    pd.to_datetime("31/03/2001", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [10],
            },
            {
                "ERROR_ID": (
                    "child8",
                    pd.to_datetime("22/06/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [11],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4017"
    assert (
        result.definition.message
        == "A CIN Plan has been reported as open at the same time as a Child Protection Plan."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
CINdetails = CINTable.CINdetails

LAchildID = ChildIdentifiers.LAchildID


@rule_definition(
    code="8590",
    module=CINTable.ChildIdentifiers,
    message="Child does not have a recorded CIN episode.",
    affected_fields=[LAchildID],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cid = data_container[ChildIdentifiers].copy()
    df_cin = data_container[CINdetails].copy()

    df_cid.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_cid.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # Each child must have at least one <CINdetails> group
    df_merge = df_cid.merge(
        df_cin[LAchildID], on=[LAchildID], how="left", indicator=True
    )

    condition = df_merge["_merge"] == "left_only"

    df_merge = df_merge[condition].reset_index()

    df_merge["ERROR_ID"] = tuple(zip(df_merge[LAchildID]))

    df_cid_issues = (
        df_cid.merge(df_merge, left_on="ROW_ID", right_on="ROW_ID")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=ChildIdentifiers, columns=[LAchildID], row_df=df_cid_issues
    )


def test_validate():
    sample_child_identifiers = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass
            },
            {
                "LAchildID": "child2",  # Pass
            },
            {
                "LAchildID": "child3",  # Fail
            },
            {
                "LAchildID": "child4",  # Pass
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass
            },
            {
                "LAchildID": "child2",  # Pass
            },
            {
                "LAchildID": "child4",  # Pass
            },
            {
                "LAchildID": "child5",  # Ignore
            },
        ]
    )

    result = run_rule(
        validate,
        {
            ChildIdentifiers: sample_child_identifiers,
            CINdetails: sample_cin_details,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 1
    issues = issues_list[0]

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns
    assert issue_columns == [LAchildID]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1

    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": ("child3",),  # ChildID
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8590"
    assert result.definition.message == "Child does not have a recorded CIN episode."



"""
Rule number: '1520'
Module: Child idenitifiers
Rule details: Each pupil <UPN> (N00001) must be unique across all pupils in the extract. 
Note: This rule should be evaluated at LA-level for imported data                                                                     
Rule message: More than one record with the same UPN
"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
UPN = ChildIdentifiers.UPN


@rule_definition(
    code="1520",
    module=CINTable.ChildIdentifiers,
    message="More than one record with the same UPN.",
    affected_fields=[UPN],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    df.index.name = "ROW_ID"

    # Each pupil <UPN> (N00001) must be unique across all pupils in the extract

    df = df[df[UPN].notna()]
    df_issues = df[df.duplicated(subset=[UPN], keep=False)].reset_index()

    link_id = tuple(
        zip(
            df_issues["LAchildID"],
            df_issues[UPN],
        )
    )

    df_issues["ERROR_ID"] = link_id

    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(table=ChildIdentifiers, columns=[UPN], row_df=df_issues)


def test_validate():
    child_identifiers = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "UPN": "1234",
            },
            {
                "LAchildID": "child2",
                "UPN": "1234",
            },
            {
                "LAchildID": "child3",
                "UPN": "12345",
            },
            {
                "LAchildID": "child4",
                "UPN": pd.NA,
            },
            {
                "LAchildID": "child4",
                "UPN": pd.NA,
            },
        ]
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    # The result contains a NamedTuple of issues encountered
    issues = result.type1_issues

    # get table name and check it. Replace ChildProtectionPlans with the name of your table.
    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    # check that the right columns were returned. Replace CPPstartDate and CPPendDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [UPN]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df

    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "1234",
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "1234",
                ),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8840 with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "1520"
    assert result.definition.message == "More than one record with the same UPN."



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
LAchildID = CINdetails.LAchildID
ReferralNFA = CINdetails.ReferralNFA
PrimaryNeedCode = CINdetails.PrimaryNeedCode
CINdetailsID = CINdetails.CINdetailsID


@rule_definition(
    code="8610",
    module=CINTable.CINdetails,
    message="Primary Need code is missing for a referral which led to further action.",
    affected_fields=[ReferralNFA, PrimaryNeedCode],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINdetails]
    df.index.name = "ROW_ID"
    df.reset_index(inplace=True)

    # If <ReferralNFA> (N00112) = false or 0
    # then
    # <PrimaryNeedCode> (N00101) must be present

    condition = (df[ReferralNFA].isin(["false", "0"])) & (df[PrimaryNeedCode].isna())
    df_issues = df[condition].reset_index()

    link_id = tuple(zip(df_issues[LAchildID], df_issues[CINdetailsID]))
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    rule_context.push_type_1(
        table=CINdetails, columns=[ReferralNFA, PrimaryNeedCode], row_df=df_issues
    )


def test_validate():
    #  Fails rows 0, 1, and 3
    sample_cin = pd.DataFrame(
        {
            "LAchildID": ["child1", "child2", "child3", "child4", "child5"],
            "ReferralNFA": [
                "false",
                "0",
                "1",
                pd.NA,
                "true",
            ],
            "PrimaryNeedCode": [
                pd.NA,  # fail
                pd.NA,  # fail
                "12/09/2022",
                "05/12/1997",
                pd.NA,  # ignore: ReferralNFA is true
            ],
            "CINdetailsID": [
                "ID1",
                "ID2",
                "ID3",
                "ID4",
                "ID5",
            ],
        }
    )

    result = run_rule(validate, {CINdetails: sample_cin})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == CINdetails
    issue_columns = issues.columns
    assert issue_columns == [ReferralNFA, PrimaryNeedCode]

    issue_rows = issues.row_df

    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "ID1",
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": ("child2", "ID2"),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8610"
    assert (
        result.definition.message
        == "Primary Need code is missing for a referral which led to further action."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildCharacteristics = CINTable.ChildCharacteristics
Ethnicity = ChildCharacteristics.Ethnicity


@rule_definition(
    code="4220",
    module=CINTable.ChildCharacteristics,
    message="Ethnicity is missing or invalid (see Ethnicity table)",
    affected_fields=[Ethnicity],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildCharacteristics]
    """
    <Ethnicity> (N00177) must be present and a valid code
    """

    eth_list = [
        "ABAN",
        "AIND",
        "AOTH",
        "APKN",
        "BAFR",
        "BCRB",
        "BOTH",
        "CHNE",
        "MOTH",
        "MWAS",
        "MWBA",
        "MWBC",
        "NOBT",
        "OOTH",
        "REFU",
        "WBRI",
        "WIRI",
        "WIRT",
        "WOTH",
        "WROM",
    ]

    df.reset_index(inplace=True)

    # Ethnicity is not in list or is null.
    df2 = df[(~df["Ethnicity"].isin(eth_list)) | df["Ethnicity"].isna()]

    failing_indices = df2.set_index("index").index

    rule_context.push_issue(
        table=ChildCharacteristics, field=Ethnicity, row=failing_indices
    )


def test_validate():
    # 0      #1      #2      #3     #4      #5      #6
    eths = ["ABAB", "AIND", "AOTH", pd.NA, "MOTH", "WOTH", "AAAA"]

    fake_dataframe = pd.DataFrame({"Ethnicity": eths})

    result = run_rule(validate, {ChildCharacteristics: fake_dataframe})

    issues = list(result.issues)

    assert len(issues) == 3

    assert issues == [
        IssueLocator(CINTable.ChildCharacteristics, Ethnicity, 0),
        IssueLocator(CINTable.ChildCharacteristics, Ethnicity, 3),
        IssueLocator(CINTable.ChildCharacteristics, Ethnicity, 6),
    ]

    assert result.definition.code == "4220"
    assert (
        result.definition.message
        == "Ethnicity is missing or invalid (see Ethnicity table)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    RuleType,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Reviews = CINTable.Reviews
CPPreviewDate = Reviews.CPPreviewDate


# define characteristics of rule
@rule_definition(
    code="8842Q",
    module=CINTable.Reviews,
    rule_type=RuleType.QUERY,
    message="Please check and either amend or provide a reason: Review Record has a missing date",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[CPPreviewDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # Replace Reviews with the name of the table you need.
    df = data_container[Reviews]

    # implement rule logic as described by the Github issue. Put the description as a comment above the implementation as shown.

    # Where a <Reviews> group is present, a valid <CPPreviewdate> (N00116) should be present within the group
    # Valid values for columns can be found in this document:
    # https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/1025195/Children_in_need_census_2022_to_2023_guide.pdf
    condition = df[CPPreviewDate].isna()

    # This is a straightforward check because it is not possible to have multiple <CPPreviewdate> that belong to the same review group.
    failing_indices = df[condition].index

    rule_context.push_issue(table=Reviews, field=CPPreviewDate, row=failing_indices)


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    reviews = pd.DataFrame(
        [["26/05/2000"], [pd.NA], ["not a date"]], columns=[CPPreviewDate]
    )

    reviews[CPPreviewDate] = pd.to_datetime(
        reviews[CPPreviewDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Reviews: reviews})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.Reviews, CPPreviewDate, 1),
        IssueLocator(CINTable.Reviews, CPPreviewDate, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    # replace 8842Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8842Q"
    assert (
        result.definition.message
        == "Please check and either amend or provide a reason: Review Record has a missing date"
    )



import numpy as np
import pandas as pd

from cin_validator.england_holidates import england_holidates


def get_values(xml_elements, table_dict: dict, xml_block):
    """
    Iterates through the input XML to extract values from XML input data for validation.
    Called in the ingress to create each table.

     :param list xml_elements: Contains elements of the collection to add to dictionary.
     :param dictionary table_dict: Dictionary containing columns of each table to get values for.
     :param DataFrame xml_block: Contains the name of the block to search the XML to make each table.
        (Tambe these are DataFrames in ingress but are they DataFrames here?)
     :returns: table_dict with XML elements where they exist, and pd.NA where they do not.
     :rtype: Dictionary
    """

    for element in xml_elements:
        try:
            table_dict[element] = xml_block.find(element).text
        except:
            table_dict[element] = pd.NA
    return table_dict


def make_date(date_input: str):
    """
    Allows Ymd or dmY date inputs, used for make_census_period.
    Important for test_validate functions.

    :param str date_input: Contains the data data to be converted to pd.datetime
        object as a string.
    :reutrns: Date data input as pd.datetime object.
    :rtype: pd.datetime object.
    """

    date = pd.to_datetime(date_input, format="%Y/%m/%d", errors="coerce")
    if pd.isna(date):
        date = pd.to_datetime(date_input, format="%d/%m/%Y", errors="coerce")
    return date


def make_census_period(reference_date: pd.Series):
    """
    Generates the census period with variables for each of the first and last
    day of the census period. Thesze are April first of the previous year and the reference date.

    :param DataFrame reference_date: A DataFrame containing data with the reference date of
        the data being validated, selected from the Header DataFrame ReferenceDate column.
    :returns: Dates of collection start and colleciton end as pd.datetime variables collection_start and
        collection_end as a tuple.
    :rtype: Tuple
    """

    # reference_date is a pandas series. Get it's value as a string by indexing the series' values array.
    reference_date = reference_date.values[0]

    # the ReferenceDate value is always the collection_end date
    collection_end = make_date(reference_date)

    # the collection start is the 1st of April of the previous year but dates from the day of the previous collection move to the next collection.
    # e.g. in the 22-23 collection, 2022-03-31 is an allowed date according to the test data.
    collection_start = make_date(reference_date) - pd.DateOffset(years=1)

    return collection_start, collection_end


def create_issue_locs(issues):
    """
    Reverses grouping of issue rows, creating a DataFrame where each row contains a single issue location.

    :param NamedTuple-like-object issues: An object containing the fields for table, columns, and
        row_df for issues found when validating data.
    :returns: DataFrame with fields for ERROR_ID, ROW_ID, columns_affected, and tables_affected for
        issues found in validation.
    :rtype: DataFrame
    """

    # expand the row_id groups such that row_id value exists per row instead of a list
    df_issue_locs = issues.row_df
    df_issue_locs = df_issue_locs.explode("ROW_ID")

    # map every row_id to its respective columns_affected list and expand that list
    df_issue_locs["columns_affected"] = df_issue_locs["ERROR_ID"].apply(
        lambda x: issues.columns
    )
    df_issue_locs = df_issue_locs.explode("columns_affected")

    # all locations from a NamedTuple object will have the same singular value of tables_affected.
    df_issue_locs["tables_affected"] = str(issues.table)[9:]

    # now a one-to-one relationship exists across table-column-row
    df_issue_locs.reset_index(inplace=True)
    df_issue_locs.drop("index", axis=1, inplace=True)

    return df_issue_locs


def process_date_columns(df: pd.DataFrame):
    """
    Takes a DataFrame in and converts all columns with Date or date in the
    title to pd.datetime objects. Used before validating rules to allow
    comparisons between dates and datetime methods.

    :param DataFrame df: DataFrame containing data to be validated.
    :returns: DataFrame with date columns formatted to pd.datetime objects.
    :rtype: DataFrame
    """

    for column in df:
        if "date" in column.lower():
            # pd.to_datetime is intelligent. It can deal with unforseen date formats
            df[column] = pd.to_datetime(df[column], errors="coerce")
    return df


def create_holidays_array():
    """
    :return numpy-object _: business day calendar object that considers the bank holiday calendar of England and Wales
    """
    return np.busdaycalendar(holidays=england_holidates)


def england_working_days(num_days: int):
    """
    This function implements a date offset based on a holiday calendar.
    :param int num_days: number of days to offset by
    :return pd.DateOffset-obj _: date offset
    """

    holiday_calendar = create_holidays_array()

    # pd.offsets.CustomBusinessDay doesn't seem to include the end date so offset by 1 so that it does.
    return pd.offsets.CustomBusinessDay(n=num_days - 1, calendar=holiday_calendar)



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

CINdetails = CINTable.CINdetails
ReferralNFA = CINdetails.ReferralNFA


@rule_definition(
    code="8568",
    module=CINTable.CINdetails,
    message="RNFA flag is missing or invalid",
    affected_fields=[ReferralNFA],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINdetails]

    # <ReferralNFA> (N00112) must be present and must be 1 or 0, true or false
    df["ReferralNFA"] = df["ReferralNFA"].astype(str).str.lower()
    df = df[~df["ReferralNFA"].isin(["1", "0", "false", "true"])]

    failing_indices = df.index

    rule_context.push_issue(table=CINdetails, field=ReferralNFA, row=failing_indices)


def test_validate():
    RNFA = [1, 0, 2, pd.NA, "true", "Woof", "Meow"]

    fake_dataframe = pd.DataFrame({"ReferralNFA": RNFA})

    result = run_rule(validate, {CINdetails: fake_dataframe})

    issues = list(result.issues)

    assert len(issues) == 4

    assert issues == [
        IssueLocator(CINTable.CINdetails, ReferralNFA, 2),
        IssueLocator(CINTable.CINdetails, ReferralNFA, 3),
        IssueLocator(CINTable.CINdetails, ReferralNFA, 5),
        IssueLocator(CINTable.CINdetails, ReferralNFA, 6),
    ]

    assert result.definition.code == "8568"
    assert result.definition.message == "RNFA flag is missing or invalid"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

CINPlanDates = CINTable.CINplanDates
CINPlanEndDate = CINPlanDates.CINPlanEndDate
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


@rule_definition(
    code="4013",
    module=CINTable.CINplanDates,
    message="CIN Plan end date must fall within the census year",
    affected_fields=[CINPlanEndDate, ReferenceDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[CINPlanDates]

    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]
    collection_start, reference_date = make_census_period(ref_date_series)

    # If <CINPlanEndDate> (N00690) is present, then<CINPlanEndDate> (N00690) must fall within [Period_of_Census] inclusive
    # A value is out of range if it is before the start or after the end.
    failing_indices = df[
        (df[CINPlanEndDate] < collection_start) | (df[CINPlanEndDate] > reference_date)
    ].index

    rule_context.push_issue(
        table=CINPlanDates, field=CINPlanEndDate, row=failing_indices
    )


def test_validate():
    fake_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]  # the census start date here will be 01/04/2021
    )
    fake_CINEndDate = pd.DataFrame(
        [
            {
                CINPlanEndDate: "01/03/2019"
            },  # 0 fail: March 1st is before April 1st, 2021. It is out of range
            {
                CINPlanEndDate: "01/04/2021"
            },  # 1 pass: April 1st is within April 1st, 2021 to March 31st, 2022.
            {
                CINPlanEndDate: "01/10/2022"
            },  # 2 fail: October 1st is after March 31st, 2022. It is out of range
        ]
    )

    fake_CINEndDate[CINPlanEndDate] = pd.to_datetime(
        fake_CINEndDate[CINPlanEndDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(validate, {CINPlanDates: fake_CINEndDate, Header: fake_header})

    issues = list(result.issues)
    assert len(issues) == 2
    assert issues == [
        # from above, index positions 0 and 2 fail.
        IssueLocator(CINTable.CINplanDates, CINPlanEndDate, 0),
        IssueLocator(CINTable.CINplanDates, CINPlanEndDate, 2),
    ]

    assert result.definition.code == "4013"
    assert (
        result.definition.message
        == "CIN Plan end date must fall within the census year"
    )



"""
Rule number: '8875'
Module: Section 47
Rule details: Where present, the <DateOfInitialCPC> (N00110) must not be a Saturday, Sunday
Rule message: The Date of Initial Child Protection Conference cannot be a weekend

"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Section47 = CINTable.Section47
DateOfInitialCPC = Section47.DateOfInitialCPC


# define characteristics of rule
@rule_definition(
    code="8875",
    module=CINTable.Section47,
    message="The Date of Initial Child Protection Conference cannot be a weekend",
    affected_fields=[DateOfInitialCPC],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[Section47]
    # <DateOfInitialCPC> should not be a Saturday, Sunday
    # .weekday() returns the integer value for each day (0-6) with weekends being 5 and 6
    failing_indices = df[df[DateOfInitialCPC].dt.weekday >= 5].index

    rule_context.push_issue(
        table=Section47, field=DateOfInitialCPC, row=failing_indices
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    child_identifiers = pd.DataFrame(
        [
            {"DateOfInitialCPC": "17/06/2021"},  # Pass
            {"DateOfInitialCPC": "09/10/2022"},  # Fail, Sunday
            {"DateOfInitialCPC": "14/03/2020"},  # Fail, Saturday
        ],
    )  # 9/10 is a Sunday and 14/3 is a Saturday

    child_identifiers[DateOfInitialCPC] = pd.to_datetime(
        child_identifiers[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(validate, {Section47: child_identifiers})

    # The result contains a list of issues encountered
    issues = list(result.issues)
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issues) == 2
    # replace the table and column name as done earlier.
    # The last numbers represent the index values where you expect the sample data to fail the validation check.
    assert issues == [
        IssueLocator(CINTable.Section47, DateOfInitialCPC, 1),
        IssueLocator(CINTable.Section47, DateOfInitialCPC, 2),
    ]

    # Check that the rule definition is what you wrote in the context above.

    assert result.definition.code == "8875"
    assert (
        result.definition.message
        == "The Date of Initial Child Protection Conference cannot be a weekend"
    )



"""
Rule number: 8555Q
Module: CIN details
Rule details: If <PersonDeathDate> (N00108) is present, then the <CINreferralDate> (N00100) must be on or before the <PersonDeathDate> (N00108)
Rule message: Child cannot be referred after its recorded date of death

"""

from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildIdentifiers = CINTable.ChildIdentifiers
PersonDeathDate = ChildIdentifiers.PersonDeathDate
LAchildID = ChildIdentifiers.LAchildID

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate
LAchildID = CINdetails.LAchildID


@rule_definition(
    code="8555Q",
    module=CINTable.CINdetails,
    rule_type=RuleType.QUERY,
    message="Child cannot be referred after its recorded date of death",
    affected_fields=[
        PersonDeathDate,
        CINreferralDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_CINDetails = data_container[CINdetails].copy()
    df_ChildIdentifiers = data_container[ChildIdentifiers].copy()

    df_CINDetails.index.name = "ROW_ID"
    df_ChildIdentifiers.index.name = "ROW_ID"

    df_CINDetails.reset_index(inplace=True)
    df_ChildIdentifiers.reset_index(inplace=True)

    # <CINreferralDate> (N00100) must be on or before the <PersonDeathDate> (N00108)

    # Remove rows with no death date
    df_ChildIdentifiers = df_ChildIdentifiers[
        df_ChildIdentifiers[PersonDeathDate].notna()
    ]

    #  Join tables
    df_merged = df_CINDetails.merge(
        df_ChildIdentifiers,
        left_on=["LAchildID"],
        right_on=["LAchildID"],
        how="left",
        suffixes=("_CINDetails", "_ChildIdentifiers"),
    )

    #  Get rows where PersonDeathDate is less than  CINreferralDate
    condition = df_merged[PersonDeathDate] < df_merged[CINreferralDate]
    df_merged = df_merged[condition].reset_index()

    # Error identifier
    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID], df_merged[CINreferralDate], df_merged[PersonDeathDate]
        )
    )
    df_CINDetails_issues = (
        df_CINDetails.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_CINDetails")
        .groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_ChildIdentifiers_issues = (
        df_ChildIdentifiers.merge(
            df_merged, left_on="ROW_ID", right_on="ROW_ID_ChildIdentifiers"
        )
        .groupby("ERROR_ID", group_keys="False")["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_CINDetails_issues
    )
    rule_context.push_type_2(
        table=ChildIdentifiers,
        columns=[PersonDeathDate],
        row_df=df_ChildIdentifiers_issues,
    )


def test_validate():
    sample_CINDetails = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINreferralDate": "26/05/2000",  # Pass as dates are the same
            },
            {
                "LAchildID": "child2",
                "CINreferralDate": "27/06/2002",  # Fails, referral after death
            },
            {
                "LAchildID": "child3",
                "CINreferralDate": "07/02/1999",  # Pass, pre death
            },
            {
                "LAchildID": "child4",
                "CINreferralDate": pd.NA,  # Ignored, no referral date
            },
            {
                "LAchildID": "child5",
                "CINreferralDate": "14/03/2001",  # Pass, dropped due to no death date
            },
        ]
    )
    sample_ChildIdentifiers = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Pass
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",  # Fails
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child3",  # Pass
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child4",  # Pass
                "PersonDeathDate": "26/05/2000",
            },
            {
                "LAchildID": "child5",  # Pass
                "PersonDeathDate": pd.NA,
            },
        ]
    )

    sample_CINDetails[CINreferralDate] = pd.to_datetime(
        sample_CINDetails[CINreferralDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_ChildIdentifiers["PersonDeathDate"] = pd.to_datetime(
        sample_ChildIdentifiers["PersonDeathDate"], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            CINdetails: sample_CINDetails,
            ChildIdentifiers: sample_ChildIdentifiers,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns
    assert issue_columns == [PersonDeathDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    # Start Date
                    pd.to_datetime("27/06/2002", format="%d/%m/%Y", errors="coerce"),
                    # Review date
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8555Q"
    assert (
        result.definition.message
        == "Child cannot be referred after its recorded date of death"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

CINplanDates = CINTable.CINplanDates
LAchildID = CINplanDates.LAchildID

CINdetails = CINTable.CINdetails
ReferralNFA = CINdetails.ReferralNFA
CINdetailsID_details = CINdetails.CINdetailsID


@rule_definition(
    code="4000",
    module=CINTable.CINplanDates,
    message="CIN Plan details provided for a referral with no further action",
    affected_fields=[
        ReferralNFA,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cpd = data_container[CINplanDates].copy()
    df_CINdetails = data_container[CINdetails].copy()

    df_cpd.index.name = "ROW_ID"
    df_CINdetails.index.name = "ROW_ID"

    df_cpd.reset_index(inplace=True)
    df_CINdetails.reset_index(inplace=True)

    df_CINdetails = df_CINdetails[
        (df_CINdetails[ReferralNFA] == "true")
        | (df_CINdetails[ReferralNFA] == 1)
        | (df_CINdetails[ReferralNFA] == "1")
        | (df_CINdetails[ReferralNFA] == True)
    ]

    df_merged = df_CINdetails.merge(
        df_cpd,
        left_on=["LAchildID", "CINdetailsID"],
        right_on=["LAchildID", "CINdetailsID"],
        how="inner",
        suffixes=("_cin", "_cpd"),
    )

    df_merged = df_merged.reset_index()

    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged["LAchildID"],
            df_merged["CINdetailsID"],
        )
    )

    df_cpp_issues = (
        df_cpd.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpd")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_CINdetails.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=CINplanDates, columns=[LAchildID], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[ReferralNFA], row_df=df_cin_issues
    )


def test_validate():
    sample_cpd = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "CINdetailsID": "CDID1",
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "CDID2",
            },
            {
                "LAchildID": "child4",
                "CINdetailsID": "CDID6",
            },
        ]
    )
    sample_cin = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # Fail, module present
                "CINdetailsID": "CDID1",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child1",  # Pass, no referralNFA
                "CINdetailsID": "CDID2",
                "ReferralNFA": pd.NA,
            },
            {
                "LAchildID": "child3",  # Pass, no module
                "CINdetailsID": "CDID6",
                "ReferralNFA": "1",
            },
            {
                "LAchildID": "child4",  # Pass, referral NFA is false
                "CINdetailsID": "CDID6",
                "ReferralNFA": "false",
            },
        ]
    )

    result = run_rule(
        validate,
        {
            CINplanDates: sample_cpd,
            CINdetails: sample_cin,
        },
    )

    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINdetails

    issue_columns = issues.columns
    assert issue_columns == [ReferralNFA]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "CDID1",
                ),
                "ROW_ID": [0],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4000"
    assert (
        result.definition.message
        == "CIN Plan details provided for a referral with no further action"
    )



import importlib
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Iterable, Optional


class CINTable(Enum):
    """
    An enumeration class (https://docs.python.org/3/library/enum.html).
    Used in validation to select CIN data modules/tables and fields/columns
    and assign them to variables. For practical reasons, this is done to
    ensure consistent spelling.
    """

    Header = Enum(
        "Header",
        [
            "Collection",
            "Year",
            "ReferenceDate",
            "SourceLevel",
            "LEA",
            "SoftwareCode",
            "Release",
            "SerialNo",
            "DateTime",
        ],
    )
    ChildIdentifiers = Enum(
        "ChildIdentifiers",
        [
            "LAchildID",
            "UPN",
            "FormerUPN",
            "UPNunknown",
            "PersonBirthDate",
            "ExpectedPersonBirthDate",
            "GenderCurrent",
            "Sex",
            "PersonDeathDate",
        ],
    )
    ChildCharacteristics = Enum(
        "ChildCharacteristics",
        [
            "LAchildID",
            "Ethnicity",
        ],
    )
    Disabilities = Enum("Disabilities", ["LAchildID", "Disability"])
    CINdetails = Enum(
        "CINdetails",
        [
            "LAchildID",
            "CINdetailsID",
            "CINreferralDate",
            "ReferralSource",
            "PrimaryNeedCode",
            "CINclosureDate",
            "ReasonForClosure",
            "DateOfInitialCPC",
            "ReferralNFA",
        ],
    )
    Assessments = Enum(
        "Assessments",
        [
            "LAchildID",
            "CINdetailsID",
            "AssessmentID",
            "AssessmentActualStartDate",
            "AssessmentInternalReviewDate",
            "AssessmentAuthorisationDate",
            "AssessmentFactors",
        ],
    )
    AssessmentFactorsList = Enum(
        "AssessmentFactorsList",
        [
            "LAchildID",
            "CINdetailsID",
            "AssessmentID",
            "AssessmentFactor",
        ],
    )
    CINplanDates = Enum(
        "CINplanDates",
        ["LAchildID", "CINdetailsID", "CINPlanStartDate", "CINPlanEndDate"],
    )
    Section47 = Enum(
        "Section47",
        [
            "LAchildID",
            "CINdetailsID",
            "S47ActualStartDate",
            "InitialCPCtarget",
            "DateOfInitialCPC",
            "ICPCnotRequired",
        ],
    )
    ChildProtectionPlans = Enum(
        "ChildProtectionPlans",
        [
            "LAchildID",
            "CINdetailsID",
            "CPPID",
            "CPPstartDate",
            "CPPendDate",
            "InitialCategoryOfAbuse",
            "LatestCategoryOfAbuse",
            "NumberOfPreviousCPP",
        ],
    )
    Reviews = Enum("Reviews", ["LAchildID", "CINdetailsID", "CPPID", "CPPreviewDate"])

    def __getattr__(self, item):
        """
        Used to get attributes within the CINtable class. Practically used to define
        fields/column variables from within tables for use in validation rules.

        :param variable item: The name of a module and field to be used for
            a validation rule.
        :returns: A variable containing a field/column for validation, or an error
            (generally on misspelling).
        :rtype: Variable, error.
        """

        if not item.startswith("_"):
            try:
                return self.value[item].name
            except KeyError as kerr:
                raise AttributeError(f"Table {self.name} has no field {item}") from kerr
        else:
            return super().__getattr__(item)


class RuleType(Enum):
    """
    An enumeration type class that defines available rule types.
    Used to assign 'Error' or 'Query' to each rule in validation.
    """

    ERROR = "Error"
    QUERY = "Query"


@dataclass(frozen=True, eq=True)
class RuleDefinition:
    """
    A dataclass type class used in each validation to assign information about
    each validation rule to the rule.

    :param int code: The rule code for each rule.
    :param function func: Used to import the validation rule function.
    :param RuleType-class rule_type: A RuleType class object accepts a string denoting if
        the rule is an error or a query.
    :param CINtable-object module: Accepts a string denoting the module/table affected by a
        validation rule.
    :param str affected_fields: The fields/columns affected by a validation rule.
    :param str message: The message to be displayed if rule is flagged.
    :returns: RuleDefinition object containing information about validation rules.
    :rtype: dataclass object.
    """

    code: str
    func: Callable
    rule_type: RuleType = RuleType.ERROR
    module: Optional[CINTable] = None
    affected_fields: Optional[Iterable[str]] = None
    message: Optional[str] = None


@dataclass(eq=True)
class YearConfig:
    deleted: list[str]
    added_or_modified: dict[str, RuleDefinition]



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

Section47 = CINTable.Section47
S47ActualStartDate = Section47.S47ActualStartDate

CINdetails = CINTable.CINdetails
CINreferralDate = CINdetails.CINreferralDate
LAchildID = CINdetails.LAchildID


@rule_definition(
    code="2889",
    module=CINTable.Section47,
    message="The S47 start date cannot be before the referral date.",
    affected_fields=[S47ActualStartDate, CINreferralDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_s47 = data_container[Section47]
    df_cin = data_container[CINdetails]

    df_s47.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    df_s47.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # Where present, the <S47ActualStartDate> (N00148) should be on or after the <CINReferralDate> (N00100)
    # Remove null S47Starts
    df_s47 = df_s47[df_s47[S47ActualStartDate].notna()]

    # Merge tables via LAchildID and CINdetailsID.
    df_merged = df_s47.merge(
        df_cin, how="left", on=["LAchildID", "CINdetailsID"], suffixes=["_47", "_cin"]
    )

    # Check for S47 Start < Cin Ref date which are the error rows
    condition = df_merged[S47ActualStartDate] < df_merged[CINreferralDate]
    df_merged = df_merged[condition].reset_index()

    df_merged["ERROR_ID"] = tuple(
        zip(
            df_merged[LAchildID],
            df_merged[S47ActualStartDate],
            df_merged[CINreferralDate],
        )
    )

    df_47_issues = (
        df_s47.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_47")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_cin.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_2(
        table=CINdetails, columns=[CINreferralDate], row_df=df_cin_issues
    )
    rule_context.push_type_2(
        table=Section47, columns=[S47ActualStartDate], row_df=df_47_issues
    )


def test_validate():
    s47_data = (
        # ID     #CINID    #S47 Date
        ("1", "45", "2020-05-05"),  # 0
        ("4", "55", "2019-04-20"),  # 1
        ("67", "66", "2014-03-21"),  # 2 fail: preceeds 2016-03-21 in #C
        ("69", "67", "2018-04-20"),  # 3
        ("69", "67", pd.NA),  # 4
        ("167", "166", "2014-03-21"),  # 5 fail: preceeds 2015-02-21 in #G
    )

    cin_data = (
        # ID     #CINID   #CIN Ref Date
        ("1", "44", "2017-05-05"),  # A
        ("4", "55", "2019-04-20"),  # B
        ("67", "66", "2016-03-21"),  # C fail
        ("67", "67", "2015-03-21"),  # D
        ("69", "67", "2018-04-20"),  # E
        ("70", "69", "2015-04-20"),  # F
        ("167", "166", "2015-02-21"),  # G fail
    )

    fake_s47 = pd.DataFrame(
        {
            "LAchildID": [x[0] for x in s47_data],
            "CINdetailsID": [x[1] for x in s47_data],
            S47ActualStartDate: [x[2] for x in s47_data],
        }
    )
    fake_cin = pd.DataFrame(
        {
            "LAchildID": [x[0] for x in cin_data],
            "CINdetailsID": [x[1] for x in cin_data],
            CINreferralDate: [x[2] for x in cin_data],
        }
    )

    fake_s47[S47ActualStartDate] = pd.to_datetime(
        fake_s47[S47ActualStartDate], format=r"%Y-%m-%d", errors="coerce"
    )
    fake_cin[CINreferralDate] = pd.to_datetime(
        fake_cin[CINreferralDate], format=r"%Y-%m-%d", errors="coerce"
    )

    result = run_rule(validate, {Section47: fake_s47, CINdetails: fake_cin})

    issues_list = result.type2_issues
    assert len(issues_list) == 2

    issues = issues_list[1]
    assert issues.table == Section47
    assert issues.columns == [S47ActualStartDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 2
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "167",  # ChildID
                    pd.to_datetime(
                        "2014-03-21", format=r"%Y-%m-%d", errors="coerce"
                    ),  # S47ActualStartDate
                    pd.to_datetime(
                        "2015-02-21", format=r"%Y-%m-%d", errors="coerce"
                    ),  # CINreferralDate
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "67",  # ChildID
                    pd.to_datetime(
                        "2014-03-21", format=r"%Y-%m-%d", errors="coerce"
                    ),  # S47ActualStartDate
                    pd.to_datetime(
                        "2016-03-21", format=r"%Y-%m-%d", errors="coerce"
                    ),  # CINreferralDate
                ),
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "2889"
    assert (
        result.definition.message
        == "The S47 start date cannot be before the referral date."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
AssessmentFactors = Assessments.AssessmentFactors
LAchildID = Assessments.LAchildID
AssessmentID = Assessments.AssessmentID

AssessmentFactorsList = CINTable.AssessmentFactorsList
AssessmentFactor = AssessmentFactorsList.AssessmentFactor

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8897Q",
    module=CINTable.Assessments,
    rule_type=RuleType.QUERY,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Parental or child factors at assessment information is missing from a completed assessment",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[AssessmentAuthorisationDate, AssessmentFactors],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    df_ass = data_container[Assessments]
    df_asslist = data_container[AssessmentFactorsList]

    header = data_container[Header]
    ref_date_series = header[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_ass.index.name = "ROW_ID"
    df_ass.reset_index(inplace=True)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # Where present, if <AssessmentAuthorisationDate> (N00160) is on or after [Start_Of_Census_Year] then one or more <AssessmentFactors>
    # (N00181) must be present within the same assessment module and must be a valid code
    # Get collection period
    factors_list = [
        "1A",
        "1B",
        "1C",
        "2A",
        "2B",
        "2C",
        "3A",
        "3B",
        "3C",
        "4A",
        "4B",
        "4C",
        "5A",
        "5B",
        "5C",
        "6A",
        "6B",
        "6C",
        "7A",
        "8B",
        "8C",
        "8D",
        "8E",
        "8F",
        "9A",
        "10A",
        "11A",
        "12A",
        "13A",
        "14A",
        "15A",
        "16A",
        "17A",
        "18A",
        "18B",
        "18C",
        "19A",
        "19B",
        "19C",
        "20",
        "21",
        "22A",
        "23A",
        "24A",
    ]

    df_ass_merged = df_ass.merge(
        df_asslist[["LAchildID", "AssessmentID", "AssessmentFactor"]],
        on=["LAchildID", "AssessmentID"],
        how="left",
    )

    condition1 = (df_ass_merged[AssessmentAuthorisationDate] >= collection_start) & (
        df_ass_merged[AssessmentAuthorisationDate].notna()
    )
    condition2 = (df_ass_merged[AssessmentFactors].notna()) & (
        df_ass_merged[AssessmentFactor].isin(factors_list)
    )

    # get all the data that fits the failing condition. Reset the index so that ROW_ID now becomes a column of df
    df_issues = df_ass_merged[condition1 & ~condition2].reset_index()

    # SUBMIT ERRORS
    # Generate a unique ID for each instance of an error. In this case,
    # - If only LAchildID is used as an identifier, multiple instances of the error on a child will be understood as 1 instance.
    # We don't want that because in reality, a child can have multiple instances of an error.
    # - If we use the LAchildID-CPPstartDate combination, that artificially cancels out the instances where a start date repeats for the same child.
    # Another rule checks for that condition. Not this one.
    # - It is very unlikely that a combination of LAchildID-CPPstartDate-CPPendDate will repeat in the DataFrame.
    # Hence, it can be used as a unique identifier of the row.

    # Replace CPPstartDate and CPPendDate below with the columns concerned in your rule.
    link_id = tuple(
        zip(
            df_issues[LAchildID],
            df_issues[AssessmentID],
            df_issues[AssessmentFactors],
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    # Ensure that you do not change the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_1(
        table=Assessments,
        columns=[AssessmentAuthorisationDate, AssessmentFactors],
        row_df=df_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    fake_data = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "AssessmentID": "11",
                "AssessmentFactors": pd.NA,
                "AssessmentAuthorisationDate": "26/05/2000",
            },  # Fails as no assessment factor code
            {
                "LAchildID": "child2",
                "AssessmentID": "21",
                "AssessmentFactors": "99",
                "AssessmentAuthorisationDate": "26/05/2000",
            },  # Fails as incorrect assessment factor code
            {
                "LAchildID": "child3",
                "AssessmentID": "31",
                "AssessmentFactors": "1A",
                "AssessmentAuthorisationDate": "26/05/2000",
            },
            {
                "LAchildID": "child3",
                "AssessmentID": "32",
                "AssessmentAuthorisationDate": "26/05/2000",
                "AssessmentFactors": pd.NA,
            },  # Fails as no factor selected
            {
                "LAchildID": "child4",
                "AssessmentID": "41",
                "AssessmentFactors": "1A",
                "AssessmentAuthorisationDate": "26/05/2000",
            },
            {
                "LAchildID": "child5",
                "AssessmentID": "51",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentFactors": pd.NA,
            },
            {
                "LAchildID": "child5",
                "AssessmentID": "52",
                "AssessmentAuthorisationDate": "26/05/1945",
                "AssessmentFactors": pd.NA,  # Passes as before census year
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.

    fake_data[AssessmentAuthorisationDate] = pd.to_datetime(
        fake_data[AssessmentAuthorisationDate], format="%d/%m/%Y", errors="coerce"
    )

    sample_header = pd.DataFrame([{ReferenceDate: "31/03/2001"}])

    # Run rule function passing in our sample data
    result = run_rule(
        validate,
        {
            Assessments: fake_data,
            AssessmentFactorsList: fake_data.rename(
                columns={"AssessmentFactors": "AssessmentFactor"}
            ),
            Header: sample_header,
        },
    )

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == Assessments

    issue_columns = issues.columns
    assert issue_columns == [AssessmentAuthorisationDate, AssessmentFactors]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 3 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 3
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",
                    "11",
                    pd.NA,
                ),
                "ROW_ID": [0],
            },
            {
                "ERROR_ID": (
                    "child2",
                    "21",
                    "99",
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "32",
                    pd.NA,
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # Check that the rule definition is what you wrote in the context above.

    # replace 8897Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8897Q"
    assert (
        result.definition.message
        == "Parental or child factors at assessment information is missing from a completed assessment"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule

ChildProtectionPlans = CINTable.ChildProtectionPlans
LAchildID = ChildProtectionPlans.LAchildID
CINdetailsID = ChildProtectionPlans.CINdetailsID
CPPID = ChildProtectionPlans.CPPID
CPPendDate = ChildProtectionPlans.CPPendDate

CINplanDates = CINTable.CINplanDates
CINPlanEndDate = CINplanDates.CINPlanEndDate
CINPlanStartDate = CINplanDates.CINPlanStartDate


@rule_definition(
    code="4001",
    module=CINTable.CINplanDates,
    message="A CIN Plan cannot run concurrently with a Child Protection Plan",
    affected_fields=[
        CINPlanEndDate,
        CPPendDate,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_cpp = data_container[ChildProtectionPlans]
    df_plan = data_container[CINplanDates]

    df_cpp.index.name = "ROW_ID"
    df_plan.index.name = "ROW_ID"

    df_cpp.reset_index(inplace=True)
    df_plan.reset_index(inplace=True)

    # If a <CINDetails> module has a <ChildProtectionPlan> module present with no <CPPendDate> (N00115)
    # - then a <CINPlanDates> module with no <CINPlanEndDate> (N00690) must not be present
    df_merged = df_cpp.merge(
        df_plan,
        on=["LAchildID", "CINdetailsID"],
        how="left",
        suffixes=("_cpp", "_cin"),
    )

    #  Get rows where CPPendDate is null and CINPlanEndDate is null
    condition = df_merged[CPPendDate].isna() & (
        df_merged[CINPlanStartDate].notna() & df_merged[CINPlanEndDate].isna()
    )
    df_merged = df_merged[condition].reset_index()

    df_merged["ERROR_ID"] = tuple(zip(df_merged[LAchildID], df_merged[CINdetailsID]))

    df_cpp_issues = (
        df_cpp.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cpp")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    df_cin_issues = (
        df_plan.merge(df_merged, left_on="ROW_ID", right_on="ROW_ID_cin")
        .groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )
    rule_context.push_type_2(
        table=ChildProtectionPlans,
        columns=[CPPendDate],
        row_df=df_cpp_issues,
    )
    rule_context.push_type_2(
        table=CINplanDates, columns=[CINPlanEndDate], row_df=df_cin_issues
    )


def test_validate():
    sample_cpp = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Pass
                "CPPendDate": "30/05/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",  # 1 Pass
                "CPPendDate": pd.NA,
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child2",  # 2 Fail
                "CPPendDate": pd.NA,
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child3",  # 3 Pass
                "CPPendDate": "30/10/2001",
                "CINdetailsID": "cinID5",
            },
            {
                "LAchildID": "child5",  # 4 Pass
                "CPPendDate": pd.NA,
                "CINdetailsID": "cinID7",
            },
            {
                "LAchildID": "child6",  # 4 Pass
                "CPPendDate": pd.NA,
                "CINdetailsID": "cinID8",
            },
        ]
    )
    sample_plan = pd.DataFrame(
        [
            {
                "LAchildID": "child1",  # 0 Pass
                "CINPlanEndDate": "04/04/2000",
                "CINPlanStartDate": "04/04/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child1",  # 1 Pass
                "CINPlanEndDate": "28/05/2000",
                "CINPlanStartDate": "04/04/2000",
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child2",  # 2 Fail
                "CINPlanStartDate": "04/04/2000",
                "CINPlanEndDate": pd.NA,
                "CINdetailsID": "cinID3",
            },
            {
                "LAchildID": "child2",  # 3 Pass
                "CINPlanStartDate": "04/04/2000",
                "CINPlanEndDate": pd.NA,
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child3",  # 4 Pass
                "CINPlanStartDate": "04/04/2000",
                "CINPlanEndDate": "30/10/2001",
                "CINdetailsID": "cinID5",
            },
            {
                "LAchildID": "child4",  # 5 Pass
                "CINPlanStartDate": "04/04/2000",
                "CINPlanEndDate": pd.NA,
                "CINdetailsID": "cinID6",
            },
            {
                "LAchildID": "child6",  # 6 Pass
                "CINPlanEndDate": pd.NA,
                "CINdetailsID": "cinID9",
            },
        ]
    )

    sample_cpp[CPPendDate] = pd.to_datetime(
        sample_cpp[CPPendDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_plan[CINPlanEndDate] = pd.to_datetime(
        sample_plan[CINPlanEndDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            CINplanDates: sample_plan,
        },
    )
    issues_list = result.type2_issues
    assert len(issues_list) == 2
    issues = issues_list[1]

    issue_table = issues.table
    assert issue_table == CINplanDates

    issue_columns = issues.columns
    assert issue_columns == [CINPlanEndDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 1
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",
                    "cinID3",
                ),
                "ROW_ID": [2],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "4001"
    assert (
        result.definition.message
        == "A CIN Plan cannot run concurrently with a Child Protection Plan"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
UPN = ChildIdentifiers.UPN


@rule_definition(
    code="1510",
    module=CINTable.ChildIdentifiers,
    message="UPN invalid (wrong check letter at character 1)",
    affected_fields=[UPN],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]

    """
    <UPN> (N00001) if present must contain the correct check letter

    To calculate the check letter:

    1. Multiply the individual digits by their weights as follows:

    digit 2 by weight 2; digit 3 by weight 3; digit 4 by weight 4; digit 5 by weight 5; digit 6 by weight 6; digit 7 by weight 7; digit 8 by weight 8; digit 9 by weight 9; digit 10 by weight 10; digit 11 by weight 11; digit 12 by weight 12; digit 13 by weight 13.

    2. Sum the individual results, divide the total by 23, and take the remainder.

    3. Calculate the check letter from the result as follows:

    0  = A;  1  = B;  2  = C;  3  = D;  4  = E;  5 = F;  6 = G;
    7 = H;  8 = J;  9 = K;  10 = L;  11 = M;  12 = N;  13 = P;
    14 = Q;  15 = R;  16 = T;  17 = U;  18 = V;  19 = W;  20 = X; Y = 21, Z = 22

    Full list avaliable here https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/807381/UPN_Guide_1.2.pdf
    """

    df.reset_index(inplace=True)

    df2 = df[["index", "UPN"]]
    df2 = df2[(df2["UPN"].str.len() == 13) & df2["UPN"].notna()]

    # the reference value is the first character of the UPN string.
    df2["FIRST_CHAR"] = df2["UPN"].str[0]

    # the last 12 characters have to be digits or else the UPN cannot be considered.
    df2["LAST_C"] = df2["UPN"].str[1:]
    df2["LAST_C"] = df2["LAST_C"].apply(lambda x: int(x) if str(x).isdigit() else pd.NA)
    df2 = df2[df2["LAST_C"].notna()]

    # calculate check value according to rule description.
    df2["SUMMED"] = 0
    for i in range(1, 13):
        # previous check was important because non-digits cannot be converted to int and hence cannot be multiplied.
        df2["SUMMED"] = (df2["SUMMED"] + (df2["UPN"].str[i].astype(int) * (i + 1))) % 23

    # enumerate object yields (index_position, element) tuples for each element in the string.
    check_map = enumerate(list("ABCDEFGHJKLMNPQRTUVWXYZ"))
    check_map = dict((i, j) for i, j in check_map)

    # Deduce the alphabet-letter representation of the calculated value based on rule description.
    df2["CHECK_CHAR"] = df2["SUMMED"].map(check_map)
    # select out all the rows where the calculated value does not match the expected value.
    df2 = df2[df2["CHECK_CHAR"].astype(str) != df2["FIRST_CHAR"].astype(str)]
    # restore the original index.
    failing_indices = df2.set_index("index").index

    rule_context.push_issue(table=ChildIdentifiers, field=UPN, row=failing_indices)


def test_validate():
    child_identifiers = pd.DataFrame(
        {
            "UPN": [
                # These should pass
                "A950000178301",  # 0 Valid format
                pd.NA,  # 1
                "H243278544154",  # 2 Valid format
                "ASFFAGSVSV123",  # 3 Nonsense
                "R325",  # 4 Nonsense
                # These should fail
                "R247962919251",  # 5 Wrong initial char
                "X428558133462",  # 6 Wrong initial char
                "X845212818005",
            ]
        }
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = list(result.issues)

    assert len(issues) == 2

    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, UPN, 5),
        IssueLocator(CINTable.ChildIdentifiers, UPN, 6),
    ]

    assert result.definition.code == "1510"
    assert (
        result.definition.message == "UPN invalid (wrong check letter at character 1)"
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import (
    CINTable,
    IssueLocator,
    RuleContext,
    rule_definition,
)
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
UPN = ChildIdentifiers.UPN


@rule_definition(
    code="1540",
    module=CINTable.ChildIdentifiers,
    message="UPN invalid (characters 5-12 not all numeric)",
    affected_fields=[UPN],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    """
    Returns indices of rows where character 5:12 of UPN contains non numerical characters.
    Does this by:
    Returning a boolean for the logic check to see is characters 5:12 contain only numerical characters.
    Using the not operator (~) to return values as false where the logic returns true (and true if there are non-numeric characters).
    Slicing df according to this criteria.
    Returns indices of the rows of this df to failing_indices."""

    df = data_container[ChildIdentifiers]

    # If <UPN> (N00001) present Characters 5-12 of <UPN> must be numeric

    #  df takes a slice of rows of df where the UPN column doesn't have Na/NaN values
    df = df.loc[df["UPN"].notna()]

    failing_indices = df[~df["UPN"].str[4:12].str.isdigit()].index

    rule_context.push_issue(table=ChildIdentifiers, field=UPN, row=failing_indices)


def test_validate():
    child_identifiers = pd.DataFrame(
        {"UPN": [pd.NA, "X000000000000", "X0000y0000000", "x0000000er00e0"]}
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = list(result.issues)
    assert len(issues) == 2
    assert issues == [
        IssueLocator(CINTable.ChildIdentifiers, UPN, 2),
        IssueLocator(CINTable.ChildIdentifiers, UPN, 3),
    ]

    assert result.definition.code == "1540"
    assert result.definition.message == "UPN invalid (characters 5-12 not all numeric)"



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import make_census_period

ChildProtectionPlans = CINTable.ChildProtectionPlans
CINdetails = CINTable.CINdetails
Section47 = CINTable.Section47

CPPstartDate = ChildProtectionPlans.CPPstartDate
LAchildID = ChildProtectionPlans.LAchildID
CINdetailsID = ChildProtectionPlans.CINdetailsID
DateOfInitialCPC = Section47.DateOfInitialCPC

# Reference date in header is needed to define the period of census.
Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    # write the rule code here, in place of '2885'
    code="2885",
    # replace ChildProtectionPlans with the value in the module column of the excel sheet corresponding to this rule .
    # Note that even if multiple tables are involved, one table will be named in the module column.
    module=CINTable.ChildProtectionPlans,
    # replace the message with the corresponding value for this rule, gotten from the excel sheet.
    message="Child protection plan shown as starting a different day to the initial child protection conference.",
    # The column names tend to be the words within the < > signs in the github issue description.
    affected_fields=[
        CPPstartDate,
        DateOfInitialCPC,
    ],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    # PREPARING DATA

    # Replace ChildProtectionPlans with the name of the table you need.
    df_cpp = data_container[ChildProtectionPlans]
    df_47 = data_container[Section47]
    df_cin = data_container[CINdetails]

    # Before you begin, rename the index so that the initial row positions can be kept intact.
    df_cpp.index.name = "ROW_ID"
    df_47.index.name = "ROW_ID"
    df_cin.index.name = "ROW_ID"

    # Resetting the index causes the ROW_IDs to become columns of their respective DataFrames
    # so that they can come along when the merge is done.

    df_cpp.reset_index(inplace=True)
    df_47.reset_index(inplace=True)
    df_cin.reset_index(inplace=True)

    # get collection period
    header = data_container[Header]
    ref_date_series = header[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    # lOGIC
    # Implement rule logic as described by the Github issue.
    # Put the description as a comment above the implementation as shown.

    # If present and if <CPPStartDate> (N00105) falls within [Period_Of_Census], then the <CPPStartDate> (N00105) should equal either:
    # a) a Section47 module <DateOfInitialCPC> (N00110), or
    # b) a CINDetails module <DateOfInitialCPC> (N00110) if there is no associated Section 47 record.

    start_date_present = df_cpp[CPPstartDate].notna()
    within_period = (df_cpp[CPPstartDate] >= collection_start) & (
        df_cpp[CPPstartDate] <= collection_end
    )
    df_cpp = df_cpp[start_date_present & within_period]

    # left merge means that only the filtered cpp children will be considered and there is no possibility of additonal children coming in from other tables.

    # get only the section47 rows where cppstartdate exists and is within period.
    df_cpp_47 = df_cpp.merge(
        df_47, on=[LAchildID, CINdetailsID], how="inner", suffixes=["_cpp", "_47"]
    )

    # FIND LOCATIONS THAT FAIL THE RULE

    # Section47 table: if CPPstartDate matches any DateOfInitialCPC within its CIN module, all DateOfInitialCPCs should pass in that CIN module.

    # locations with equal dates surely pass.
    df_cpp_47_pass = df_cpp_47[df_cpp_47[CPPstartDate] == df_cpp_47[DateOfInitialCPC]]
    # locations with unequal dates could fail.
    df_cpp_47_failable = df_cpp_47[
        df_cpp_47[CPPstartDate] != df_cpp_47[DateOfInitialCPC]
    ]
    # the failable locations that are not found in a CIN module where a DateOfInitialCPC passes, surely fail.
    df_cpp_47_failable["ERROR_ID"] = tuple(
        zip(df_cpp_47_failable[LAchildID], df_cpp_47_failable[CINdetailsID])
    )
    df_cpp_47_pass["ERROR_ID"] = tuple(
        zip(df_cpp_47_pass[LAchildID], df_cpp_47_pass[CINdetailsID])
    )
    df_cpp_47_fail = df_cpp_47_failable[
        ~(df_cpp_47_failable["ERROR_ID"].isin(df_cpp_47_pass["ERROR_ID"]))
    ]

    # One CPPstartDate's pass should not affect the other. Create a separate reference dataset for CPPstartDate failing locations.
    # Redo the above steps to create a dataset where a CPPstartDate is only removed from failable if its LAchildID-CINdetails-CPPstartDate is found to pass.
    df_cpp_47_failable["ERROR_startdate"] = tuple(
        zip(
            df_cpp_47_failable[LAchildID],
            df_cpp_47_failable[CINdetailsID],
            df_cpp_47_failable[CPPstartDate],
        )
    )
    df_cpp_47_pass["ERROR_startdate"] = tuple(
        zip(
            df_cpp_47_pass[LAchildID],
            df_cpp_47_pass[CINdetailsID],
            df_cpp_47_pass[CPPstartDate],
        )
    )
    df_cpp_47_startdate_fail = df_cpp_47_failable[
        ~(df_cpp_47_failable["ERROR_startdate"].isin(df_cpp_47_pass["ERROR_startdate"]))
    ]

    # CIN table: if CPPstartDate matches any DateOfInitialCPC within its CIN module, all DateOfInitialCPCs should pass in that CIN module.

    df_cpp_cin = df_cpp.merge(
        df_cin, on=[LAchildID, CINdetailsID], how="left", suffixes=["_cpp", "_cin"]
    )
    df_cpp_cin_pass = df_cpp_cin[
        df_cpp_cin[CPPstartDate] == df_cpp_cin[DateOfInitialCPC]
    ]
    df_cpp_cin_failable = df_cpp_cin[
        df_cpp_cin[CPPstartDate] != df_cpp_cin[DateOfInitialCPC]
    ]

    df_cpp_cin_failable["ERROR_ID"] = tuple(
        zip(df_cpp_cin_failable[LAchildID], df_cpp_cin_failable[CINdetailsID])
    )
    df_cpp_cin_pass["ERROR_ID"] = tuple(
        zip(df_cpp_cin_pass[LAchildID], df_cpp_cin_pass[CINdetailsID])
    )
    df_cpp_cin_fail = df_cpp_cin_failable[
        ~(df_cpp_cin_failable["ERROR_ID"].isin(df_cpp_cin_pass["ERROR_ID"]))
    ]

    df_cpp_cin_failable["ERROR_startdate"] = tuple(
        zip(
            df_cpp_cin_failable[LAchildID],
            df_cpp_cin_failable[CINdetailsID],
            df_cpp_cin_failable[CPPstartDate],
        )
    )
    df_cpp_cin_pass["ERROR_startdate"] = tuple(
        zip(
            df_cpp_cin_pass[LAchildID],
            df_cpp_cin_pass[CINdetailsID],
            df_cpp_cin_pass[CPPstartDate],
        )
    )
    df_cpp_cin_startdate_fail = df_cpp_cin_failable[
        ~(
            df_cpp_cin_failable["ERROR_startdate"].isin(
                df_cpp_cin_pass["ERROR_startdate"]
            )
        )
    ]

    # Since the condition is to check the CIN table only if the LAchildID-CINdetailID is absent in Section47,
    # only those failing locations that are absent in section47 table should be flagged in the cindetails table. Locations that are flagged in section47 don't need to re-flag here.

    # create an ID that cin groups can be identified by.
    df_47["ERROR_ID"] = tuple(zip(df_47[LAchildID], df_47[CINdetailsID]))
    df_cpp_cin_fail_no_47 = df_cpp_cin_fail[
        ~(df_cpp_cin_fail["ERROR_ID"].isin(df_47["ERROR_ID"]))
    ]
    # do same for the cppstartdate identifier table.
    df_cpp_cin_startdate_fail_no47 = df_cpp_cin_startdate_fail[
        ~(df_cpp_cin_startdate_fail["ERROR_ID"].isin(df_47["ERROR_ID"]))
    ]

    # MAP FAILING LOCATIONS TO THEIR ORIGINAL TABLES.
    df_47_issues = df_47[df_47["ROW_ID"].isin(df_cpp_47_fail["ROW_ID_47"])]
    df_47_issues = (
        df_47_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cin["ERROR_ID"] = tuple(zip(df_cin[LAchildID], df_cin[CINdetailsID]))
    df_cin_issues = df_cin[df_cin["ROW_ID"].isin(df_cpp_cin_fail_no_47["ROW_ID_cin"])]
    df_cin_issues = (
        df_cin_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    df_cpp["ERROR_ID"] = tuple(
        zip(df_cpp[LAchildID], df_cpp[CINdetailsID], df_cpp[CPPstartDate])
    )
    df_cpp_issues = df_cpp[
        (df_cpp["ROW_ID"].isin(df_cpp_47_startdate_fail["ROW_ID_cpp"]))
        | (df_cpp["ROW_ID"].isin(df_cpp_cin_startdate_fail_no47["ROW_ID_cpp"]))
    ]
    df_cpp_issues = (
        df_cpp_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    # In this case, the rule is checked for each CPPstartDate, in each CINplanDates group (differentiated by CINdetailsID), in each child (differentiated by LAchildID)
    # However, the cin and section47 locations fail/pass per group so LAchildID-CINdetailsID is used to identify them.
    # On the other hand, CPPstartDate locations fail/pass independently so a combination of LAchildID, CINdetailsID and CPPstartDate identifies a cpp error instance.

    # Ensure that you maintain the ROW_ID, and ERROR_ID column names which are shown above. They are keywords in this project.
    rule_context.push_type_2(
        table=ChildProtectionPlans, columns=[CPPstartDate], row_df=df_cpp_issues
    )
    rule_context.push_type_2(
        table=Section47, columns=[DateOfInitialCPC], row_df=df_47_issues
    )
    rule_context.push_type_2(
        table=CINdetails, columns=[DateOfInitialCPC], row_df=df_cin_issues
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_cpp = pd.DataFrame(
        [  # child1: Simulates multiple cin modules with one out of census period, and multiple CPPs within the same CIN.
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID1",
                "CPPstartDate": "26/05/2021",  #  passes in section47
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID1",  # simulates multiple CPPs with same LAchildID-CINdetailsID where some fail, some pass.
                "CPPstartDate": "26/06/2021",  #  # fail. fails both section47 and cin
            },
            {
                "LAchildID": "child1",
                "CINdetailsID": "cinID2",
                "CPPstartDate": "27/06/2002",  # ignore. would've failed but ignored. Not in period of census
            },
            # child2: fail in section47, pass in cin
            {
                "LAchildID": "child2",
                "CINdetailsID": "cinID1",
                "CPPstartDate": "26/05/2021",  # fail. fails in section47, pass in cin not considered.
            },
            # child3: multiple cin details modules.
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID1",
                "CPPstartDate": "26/05/2021",  # fail. fails in both section47 and cin
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID2",
                "CPPstartDate": pd.NA,  # ignore. cppstartdate is absent
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID3",
                "CPPstartDate": "07/02/2022",  # fail. fails both section47 and cin
            },
            {
                "LAchildID": "child3",
                "CINdetailsID": "cinID4",
                "CPPstartDate": "14/03/2022",  # fail. fails in section47, pass in cin not considered.
            },
            # child 5: date present in section47 and absent in cin
            {
                "LAchildID": "child5",
                "CINdetailsID": "cinID4",
                "CPPstartDate": "19/07/2021",  # passes in section47
            },
            # child 6: no DateOfInitialCPC recorded in cin or section47 table
            {
                "LAchildID": "child6",
                "CINdetailsID": "cinID4",
                "CPPstartDate": "19/07/2021",  # fail
            },
            # child 8: multiple section47s in the same cin module where some pass and others fail.
            {
                "LAchildID": "child8",
                "CINdetailsID": "cinID1",
                "CPPstartDate": "20/10/2021",  # passes in section_47
            },
            # child 9: present in cin, absent in section47
            {
                "LAchildID": "child9",
                "CINdetailsID": "cinID1",
                "CPPstartDate": "20/10/2021",  # passes in cin
            },
        ]
    )
    sample_section47 = pd.DataFrame(
        [
            {  # 0 pass
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2021",  # pass. same as cppstartdate
                "CINdetailsID": "cinID1",
            },
            {  # 1 ignored
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2021",  # ignore. cppstartdate not in period of census
                "CINdetailsID": "cinID2",
            },
            {  # 2 pass
                "LAchildID": "child2",
                "DateOfInitialCPC": "30/05/2021",  # fail. not the same
                "CINdetailsID": "cinID1",
            },
            {  # 4 absent, ignored
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2021",  # ignore. cppstartdate is absent.
                "CINdetailsID": "cinID2",
            },
            {  # 5 fail
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2021",  # fail. not the same
                "CINdetailsID": "cinID3",
            },
            {  # 6 pass
                "LAchildID": "child3",
                "DateOfInitialCPC": pd.NA,  # fail. not the same
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child5",
                "DateOfInitialCPC": "19/07/2021",  # pass. same as cppstartdate
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child5",
                "DateOfInitialCPC": pd.NA,  # pass since other section47 in same modeule passes.
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child6",
                "DateOfInitialCPC": pd.NA,  # fail. not the same
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child8",
                "DateOfInitialCPC": "20/10/2021",  # pass. same as cpp_start_date
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child8",
                "DateOfInitialCPC": "22/07/2021",  # pass since other section47 in the same CINmodule passes.
                "CINdetailsID": "cinID1",
            },
        ]
    )
    sample_cin_details = pd.DataFrame(
        [
            {  # 0 pass
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/10/2020",  # ignore fail. not the same but present in section47 table
                "CINdetailsID": "cinID1",
            },
            {  # 1 ignore
                "LAchildID": "child1",
                "DateOfInitialCPC": "26/05/2021",  # ignore. cppstartdate not in period of census
                "CINdetailsID": "cinID2",
            },
            {  # 2 pass
                "LAchildID": "child2",
                "DateOfInitialCPC": "26/05/2021",  # ignore fail. could've passed but present in section47
                "CINdetailsID": "cinID1",
            },
            {  # 3 fail
                "LAchildID": "child3",
                "DateOfInitialCPC": "28/05/2021",  # fail. not the same and no corresponding section47
                "CINdetailsID": "cinID1",
            },
            {  # 4 ignore
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2021",  # ignore. cppstartdate is absent
                "CINdetailsID": "cinID2",
            },
            {  # 5 fail
                "LAchildID": "child3",
                "DateOfInitialCPC": "26/05/2003",  # ignore fail. not the same and has corresponding section47.
                "CINdetailsID": "cinID3",
            },
            {  # 6 pass
                "LAchildID": "child3",
                "DateOfInitialCPC": "14/03/2022",  # ignore fail. could've passed but present in section47
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child5",
                "DateOfInitialCPC": pd.NA,  # ignore fail. not the same  and has corresponding section47.
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child6",
                "DateOfInitialCPC": pd.NA,  # ignore fail. not the same and has corresponding section47
                "CINdetailsID": "cinID4",
            },
            {
                "LAchildID": "child7",
                "DateOfInitialCPC": pd.NA,  # ignore. not present in cpp table.
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child8",
                "DateOfInitialCPC": pd.NA,  # passes in section47
                "CINdetailsID": "cinID1",
            },
            {
                "LAchildID": "child9",
                "CINdetailsID": "cinID1",
                "DateOfInitialCPC": "20/10/2021",  # passes in cin
            },
        ]
    )
    # if rule requires columns containing date values, convert those columns to datetime objects first. Do it here in the test_validate function, not above.
    sample_header = pd.DataFrame(
        [{ReferenceDate: "31/03/2022"}]  # the census start date here will be 01/04/2021
    )
    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_cpp[CPPstartDate] = pd.to_datetime(
        sample_cpp[CPPstartDate], format="%d/%m/%Y", errors="coerce"
    )
    sample_section47[DateOfInitialCPC] = pd.to_datetime(
        sample_section47[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )
    sample_cin_details[DateOfInitialCPC] = pd.to_datetime(
        sample_cin_details[DateOfInitialCPC], format="%d/%m/%Y", errors="coerce"
    )

    # Run the rule function, passing in our sample data.
    result = run_rule(
        validate,
        {
            ChildProtectionPlans: sample_cpp,
            Section47: sample_section47,
            CINdetails: sample_cin_details,
            Header: sample_header,
        },
    )

    # Use .type2_issues to check for the result of .push_type2_issues() which you used above.
    issues_list = result.type2_issues
    assert len(issues_list) == 3
    # the function returns a list on NamedTuples where each NamedTuple contains (table, column_list, df_issues)

    # pick any table and check it's values. the tuple in location 1 will contain the Section47 columns because that's the second thing pushed above.
    issues = issues_list[1]

    # get table name and check it. Replace Section47 with the name of your table.
    issue_table = issues.table
    assert issue_table == Section47

    # check that the right columns were returned. Replace DateOfInitialCPC  with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [DateOfInitialCPC]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 2 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 4
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    # - The first column is ERROR_ID which contains the unique combination that identifies each error instance, which you decided on, in your zip, earlier.
    # - The second column in ROW_ID which contains a list of index positions that belong to each error instance.

    # The ROW ID values represent the index positions where you expect the sample data to fail the validation check.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    "cinID1",  # CINdetailsID,
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID3",
                ),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": (
                    "child3",
                    "cinID4",
                ),
                "ROW_ID": [5],
            },
            {
                "ERROR_ID": (
                    "child6",
                    "cinID4",
                ),
                "ROW_ID": [8],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    # check table 2
    issues2 = issues_list[2]
    issue_table2 = issues2.table
    assert issue_table2 == CINdetails

    issue_columns2 = issues2.columns
    assert issue_columns2 == [DateOfInitialCPC]

    issue_rows2 = issues2.row_df
    assert len(issue_rows2) == 1
    assert isinstance(issue_rows2, pd.DataFrame)
    assert issue_rows2.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df2 = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID1",  # CINdetailsID,
                ),
                "ROW_ID": [3],
            },
        ]
    )
    assert issue_rows2.equals(expected_df2)

    # check table 0
    issues0 = issues_list[0]
    issue_table0 = issues0.table
    assert issue_table0 == ChildProtectionPlans

    issue_columns0 = issues0.columns
    assert issue_columns0 == [CPPstartDate]

    issue_rows0 = issues0.row_df
    assert len(issue_rows0) == 6
    assert isinstance(issue_rows0, pd.DataFrame)
    assert issue_rows0.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df0 = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child1",  # ChildID
                    "cinID1",  # CINdetailsID,
                    pd.to_datetime("26/06/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "child2",  # ChildID
                    "cinID1",  # CINdetailsID,
                    pd.to_datetime("26/05/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID1",  # CINdetailsID,
                    pd.to_datetime("26/05/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID3",  # CINdetailsID,
                    pd.to_datetime("07/02/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [6],
            },
            {
                "ERROR_ID": (
                    "child3",  # ChildID
                    "cinID4",  # CINdetailsID,
                    pd.to_datetime("14/03/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [7],
            },
            {
                "ERROR_ID": (
                    "child6",  # ChildID
                    "cinID4",  # CINdetailsID,
                    pd.to_datetime("19/07/2021", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [9],
            },
        ]
    )
    assert issue_rows0.equals(expected_df0)

    # Confirm that the rule details were properly pushed through.
    assert result.definition.code == "2885"
    assert (
        result.definition.message
        == "Child protection plan shown as starting a different day to the initial child protection conference."
    )



from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule

ChildIdentifiers = CINTable.ChildIdentifiers
PersonBirthDate = ChildIdentifiers.PersonBirthDate
PersonDeathDate = ChildIdentifiers.PersonDeathDate
LAchildID = ChildIdentifiers.LAchildID


@rule_definition(
    code="8535Q",
    module=CINTable.ChildIdentifiers,
    rule_type=RuleType.QUERY,
    message="Please check and either amend data or provide a reason: Child’s date of death should not be prior to the date of birth",
    affected_fields=[PersonDeathDate, PersonBirthDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df = data_container[ChildIdentifiers]
    df.index.name = "ROW_ID"

    # <PersonDeathDate> (N00108) must be on or after <PersonBirthDate> (N00066)

    # Remove all rows with no deathdate
    df = df[~df[PersonDeathDate].isna()]

    # Return rows where DOB is prior to DOD
    condition1 = df[PersonBirthDate] > df[PersonDeathDate]
    # Return rows with no DOB
    condition2 = df[PersonBirthDate].isna()

    # df with all rows meeting the conditions
    df_issues = df[condition1 | condition2].reset_index()

    link_id = tuple(
        zip(
            df_issues[LAchildID], df_issues[PersonDeathDate], df_issues[PersonBirthDate]
        )
    )
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=ChildIdentifiers,
        columns=[PersonDeathDate, PersonBirthDate],
        row_df=df_issues,
    )


def test_validate():
    child_identifiers = pd.DataFrame(
        [
            {
                "LAchildID": "child1",
                "PersonDeathDate": "26/05/2000",
                "PersonBirthDate": "26/05/2000",
            },
            {
                "LAchildID": "child2",
                "PersonDeathDate": "26/05/2001",
                "PersonBirthDate": "26/05/2000",
            },
            {
                "LAchildID": "child3",
                "PersonDeathDate": "26/05/1999",
                "PersonBirthDate": "26/05/2000",
            },  # 2 error: end is before start
            {
                "LAchildID": "child4",
                "PersonDeathDate": "26/05/2000",
                "PersonBirthDate": pd.NA,
                # 3 fail: no birth date
            },
            {
                "LAchildID": "child5",
                "PersonDeathDate": "25/05/2000",
                "PersonBirthDate": "26/05/2000",
            },  # 4 error: end is before start
            {
                "LAchildID": "child6",
                "PersonDeathDate": pd.NA,
                "PersonBirthDate": pd.NA,
            },
        ]
    )

    child_identifiers[PersonDeathDate] = pd.to_datetime(
        child_identifiers[PersonDeathDate], format="%d/%m/%Y", errors="coerce"
    )
    child_identifiers[PersonBirthDate] = pd.to_datetime(
        child_identifiers[PersonBirthDate], format="%d/%m/%Y", errors="coerce"
    )

    result = run_rule(validate, {ChildIdentifiers: child_identifiers})

    issues = result.type1_issues

    issue_table = issues.table
    assert issue_table == ChildIdentifiers

    issue_columns = issues.columns
    assert issue_columns == [PersonDeathDate, PersonBirthDate]

    issue_rows = issues.row_df
    assert len(issue_rows) == 3
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "child3",
                    pd.to_datetime("26/05/1999", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [2],
            },
            {
                "ERROR_ID": (
                    "child4",
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime(pd.NA, format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
            {
                "ERROR_ID": (
                    "child5",
                    pd.to_datetime("25/05/2000", format="%d/%m/%Y", errors="coerce"),
                    pd.to_datetime("26/05/2000", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [4],
            },
        ]
    )
    assert issue_rows.equals(expected_df)

    assert result.definition.code == "8535Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: Child’s date of death should not be prior to the date of birth"
    )



"""
Rule number: 8670Q
Module: Assessments
Rule details: Where present, within an <Assessments> group, if <AssessmentAuthorisationDate> (N00160) is not present then <AssessmentActualStartDate>
(N00159) should not be before the <ReferenceDate> (N00603) minus 45 working days.

Rule message: Please check: Assessment started more than 45 working days before the end of the census year. However, there is no Assessment end date. 
"""
from typing import Mapping

import pandas as pd

from cin_validator.rule_engine import CINTable, RuleContext, RuleType, rule_definition
from cin_validator.test_engine import run_rule
from cin_validator.utils import england_working_days, make_census_period

# Get tables and columns of interest from the CINTable object defined in rule_engine/__api.py

Assessments = CINTable.Assessments
AssessmentAuthorisationDate = Assessments.AssessmentAuthorisationDate
AssessmentActualStartDate = Assessments.AssessmentActualStartDate
LAchildID = Assessments.LAchildID

Header = CINTable.Header
ReferenceDate = Header.ReferenceDate


# define characteristics of rule
@rule_definition(
    code="8670Q",
    module=CINTable.Assessments,
    rule_type=RuleType.QUERY,
    message="Please check and either amend data or provide a reason: Assessment started more than 45 working days before the end of the census year. However, there is no Assessment end date.",
    affected_fields=[AssessmentActualStartDate],
)
def validate(
    data_container: Mapping[CINTable, pd.DataFrame], rule_context: RuleContext
):
    df_assessments = data_container[Assessments]
    df_assessments.index.name = "ROW_ID"

    df_ref = data_container[Header]
    ref_date_series = df_ref[ReferenceDate]
    collection_start, collection_end = make_census_period(ref_date_series)

    #  If <AssessmentAuthorisationDate> (N00160) is not present then <AssessmentActualStartDate> (N00159) should not be before the <ReferenceDate> (N00603) minus 45 working days

    # Filter to only those with no authorisation date
    df_assessments = df_assessments[df_assessments[AssessmentAuthorisationDate].isna()]

    latest_date = collection_end - england_working_days(45)
    df_issues = df_assessments[
        df_assessments[AssessmentActualStartDate] < latest_date
    ].reset_index()

    link_id = tuple(zip(df_issues[LAchildID], df_issues[AssessmentActualStartDate]))
    df_issues["ERROR_ID"] = link_id
    df_issues = (
        df_issues.groupby("ERROR_ID", group_keys=False)["ROW_ID"]
        .apply(list)
        .reset_index()
    )

    rule_context.push_type_1(
        table=Assessments,
        columns=[AssessmentAuthorisationDate, AssessmentActualStartDate],
        row_df=df_issues,
    )


def test_validate():
    # Create some sample data such that some values pass the validation and some fail.
    sample_assessments = pd.DataFrame(
        [
            {
                "LAchildID": "ID1",
                "AssessmentAuthorisationDate": "26/05/2022",
                "AssessmentActualStartDate": "15/10/2022",
                # Pass, authorisation date present
            },
            {
                "LAchildID": "ID2",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "15/10/2022",
                # Fail, start date is before ref date - 45wd
            },
            {
                "LAchildID": "ID3",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "15/03/2023",
                # Pass, start date is withing 45wd of reference date
            },
            {
                "LAchildID": "ID4",
                "AssessmentAuthorisationDate": pd.NA,
                "AssessmentActualStartDate": "29/01/2023",
                # Fail, start date is outside of 45wd of reference date
            },
        ]
    )

    sample_assessments[AssessmentAuthorisationDate] = pd.to_datetime(
        sample_assessments[AssessmentAuthorisationDate],
        format="%d/%m/%Y",
        errors="coerce",
    )
    sample_assessments[AssessmentActualStartDate] = pd.to_datetime(
        sample_assessments[AssessmentActualStartDate],
        format="%d/%m/%Y",
        errors="coerce",
    )

    sample_header = pd.DataFrame([{"ReferenceDate": "31/03/2023"}])

    sample_header[ReferenceDate] = pd.to_datetime(
        sample_header[ReferenceDate], format="%d/%m/%Y", errors="coerce"
    )

    # Run rule function passing in our sample data
    result = run_rule(
        validate, {Assessments: sample_assessments, Header: sample_header}
    )

    # Use .type1_issues to check for the result of .push_type1_issues() which you used above.
    issues = result.type1_issues

    # get table name and check it. Replace Assessments with the name of your table.
    issue_table = issues.table
    assert issue_table == Assessments

    # check that the right columns were returned. Replace AssessmentAuthorisationDate and AssessmentActualStartDate with a list of your columns.
    issue_columns = issues.columns
    assert issue_columns == [AssessmentAuthorisationDate, AssessmentActualStartDate]

    # check that the location linking dataframe was formed properly.
    issue_rows = issues.row_df
    # replace 1 with the number of failing points you expect from the sample data.
    assert len(issue_rows) == 2
    # check that the failing locations are contained in a DataFrame having the appropriate columns. These lines do not change.
    assert isinstance(issue_rows, pd.DataFrame)
    assert issue_rows.columns.to_list() == ["ERROR_ID", "ROW_ID"]

    # Create the dataframe which you expect, based on the fake data you created. It should have two columns.
    expected_df = pd.DataFrame(
        [
            {
                "ERROR_ID": (
                    "ID2",
                    pd.to_datetime("15/10/2022", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [1],
            },
            {
                "ERROR_ID": (
                    "ID4",
                    pd.to_datetime("29/01/2023", format="%d/%m/%Y", errors="coerce"),
                ),
                "ROW_ID": [3],
            },
        ]
    )

    assert issue_rows.equals(expected_df)

    # replace 8670Q with the rule code and put the appropriate message in its place too.
    assert result.definition.code == "8670Q"
    assert (
        result.definition.message
        == "Please check and either amend data or provide a reason: Assessment started more than 45 working days before the end of the census year. However, there is no Assessment end date."
    )
